/*
* Slice
* v 0.2 (2006-12-07)
* File splitter
*
* Copyright 2006 Zach Scrivena
* Email: zachscrivena@gmail.com
* Webpage: http://zs.freeshell.org/
*
* Splits a file into several 'slices' that can be easily concatenated
* to recover the original file. Free software written in C.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
* Compile notes:
* This source file has been compiled succesfully using
* cygwin's GCC 3.4.4
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>

#define BUFFER_SIZE 1048576 // 1 Mb
#define STRING_SIZE 256

/* prints a given long long int as a comma-grouped numeral */
void printnum(long long l)
{
	char a[STRING_SIZE];
	char s[STRING_SIZE];
	int i;
	int j;
	int n;

	sprintf(a, "%lld", l);
	n = strlen(a);

	j = 0;
	for (i = 0; i < n; i ++)
	{
		s[j] = a[i];
		j ++;

		if (((n-i) % 3 == 1) && (i < n - 1))
		{
			s[j] = ','; // insert a comma
			j ++;
		}
	}

	s[j] = '\0'; // null-terminate the string
	printf("%s", s);
}

int main(int argc, char *argv[])
{

	int fd_in;    // input file descriptor (just for getting file size)
	FILE *fp_in;  // input file pointer
	FILE *fp_out; // output file pointer
	char *command; // first command-line argument ("command")
	char *file_in;  // second command-line argument ("file")
	char *file_out; // output file name
	long long commandInteger; // integer associated with the "command"
	long long size;    // input file size in bytes
	long long sliceSize;     // size of each slice (except the last slice) in bytes
	long long lastSliceSize; // size of the last slice in bytes
	long long numSlices; // number of slices
	long long sliceIndex;
	long long thisSliceSize;
	long long bytesWritten;
	int numDigits;
	int bytesToWrite;
	int bytesRead;
	char a[STRING_SIZE]; // temp var
	char formatString[STRING_SIZE];
	char buffer[BUFFER_SIZE];
	struct stat statBuffer;

	printf("\nSlice 0.2   Copyright 2006 Zach Scrivena   2006-12-07");

	if (argc == 1)
	{
		// print usage
		printf("\n\nSplits a file into several 'slices' that can be easily concatenated\nto recover the original file.");
		printf("\n\nUSAGE:   slice  [command]  [file]");
		printf("\n\nwhere\n");
		printf("\n [command] is a positive integer followed possibly by the character 'x'.");
		printf("\n   With the 'x', the integer gives the total number of slices; without the 'x',");
		printf("\n   the integer gives the maximum size of each slice in bytes.\n");
		printf("\n [file] is the name of the file to be split.\n");
		printf("\nEXAMPLES:");
		printf("\n  slice 5x work.dat        Split the file into 5 slices.");
		printf("\n  slice 1000000 work.dat   Split the file into slices that are at most");
		printf("\n                           1000000 bytes each.");
		printf("\n\n");
		exit(0);
	}
	else if (argc != 3)
	{
		printf("\n\nERROR: Invalid number of arguments.\nTo display help, run Slice without any command-line arguments.\n");
		exit(1);
	}

	// correct number of arguments supplied
	command = argv[1];
	file_in = argv[2];

	printf("\n\nFile to be sliced: %s ", file_in);

	// open file for binary read-only, to determine file size
	if ((fd_in = open(file_in, O_RDONLY)) == 0)
	{
		printf("\n\nERROR: Unable to open file for reading.\n", file_in);
		exit(1);
	}

	fstat(fd_in, &statBuffer);
	size = (long long) statBuffer.st_size;
	printf("(");
	printnum(size);
	printf(" bytes)");
	close(fd_in);

	// open the input file for binary read-only
	if ((fp_in = fopen(file_in, "rb")) == 0)
	{
		printf("\n\nERROR: Unable to open file for reading.\n", file_in);
		exit(1);
	}

	// extract the integer part of the command
	commandInteger = strtoll(command, (char **) NULL, 10);

	if (commandInteger <= 0)
	{
		printf("\n\nERROR: Invalid command %s.\nTo display help, run Slice without any command-line arguments.\n", command);
		exit(1);
	}

	// compute slice sizes
	if (command[strlen(command) - 1] == 'x' ||
	    command[strlen(command) - 1] == 'X')
	{
		// split file into specified number of slices
		sliceSize = size / commandInteger;
		numSlices = commandInteger;
	}
	else
	{
		// split file into slices not exceeding specified size in bytes
		sliceSize = commandInteger;
		numSlices = (size + commandInteger - 1) / commandInteger;
	}

	if (numSlices < 1 ||
	    sliceSize < 1)
	{
		printf("\n\nERROR: Unable to slice file to given specification.\nTry relaxing the specification.\n");
		exit(0);
	}

	if (numSlices == 1)
	{
		printf("\n\nNothing to do; original file already satisfies specification.\n");
		exit(0);
	}

	if (sliceSize > size)
	{
		sliceSize = size;
	}

	// compute size of the last slice
	lastSliceSize = size - (numSlices - 1) * sliceSize;

	// display size of slices
	printf("\nCreating ");
	printnum(numSlices);
	printf(" slices ");

	if (lastSliceSize == sliceSize)
	{
		printf("(");
		printnum(numSlices);
		printf(" x ");
		printnum(sliceSize);
		printf(" bytes)");
	}
	else
	{
		printf("(");
		printnum(numSlices - 1);
		printf(" x ");
		printnum(sliceSize);
		printf(" bytes + 1 x ");
		printnum(lastSliceSize);
		printf(" bytes)");
	}

	fflush(stdout);

	// compute number of digits required to display the maximum slice index
	numDigits = sprintf(a, "%lld", numSlices);

	// create format string used in creating the file extension of each slice
	sprintf(formatString, "%%0%dd", numDigits);

	// allocate memory for the name of the output files
	if ((file_out = (char *) malloc((strlen(file_in) + 1 + numDigits) * sizeof(char))) == NULL)
	{
		printf("\n\nERROR: Insufficient memory to create the name of the output files.");
		exit(1);
	}

	for (sliceIndex = 0; sliceIndex < numSlices; sliceIndex++)
	{
		// create the name of the output file, e.g. work.dat.001
		strcpy(file_out, file_in);
		strcat(file_out, ".");
		sprintf(a, formatString, sliceIndex + 1);
		strcat(file_out, a);

		// open output file for binary write-only
		if ((fp_out = fopen(file_out, "wb")) == NULL)
		{
			printf("\n\nERROR: Unable to create output file %s.", file_out);
			exit(1);
		}

		// compute size of the current slice (output file)
		if (sliceIndex == numSlices - 1)
		{
			// last slice
			thisSliceSize = lastSliceSize;
		}
		else
		{
			// not the last slice yet
			thisSliceSize = sliceSize;
		}

		// Read and write the current slice:

		// display a progress indicator
		printf(".");
		fflush(stdout);

		// reset total bytes written for the current slice
		bytesWritten = 0;

		while (bytesWritten < thisSliceSize)
		{
			if (thisSliceSize - bytesWritten < BUFFER_SIZE)
			{
				bytesToWrite = (int) (thisSliceSize - bytesWritten);
			}
			else
			{
				bytesToWrite = BUFFER_SIZE;
			}

			// read bytes from input file into the buffer
			bytesRead = fread(buffer, 1, bytesToWrite, fp_in);

			if (bytesRead != bytesToWrite)
			{
				printf("\n\nERROR: Unable to read input file.");
				exit(1);
			}

			// write bytes from buffer to the current slice (output file)
			fwrite(buffer, 1, bytesRead, fp_out);

			// increment total bytes written for the current slice
			bytesWritten += bytesToWrite;
		}

		// close the current slice (output file)
		fclose(fp_out);

	}

	// close the input file
	fclose(fp_in);

	// free previously allocated memory
	free(file_out);

	// successful termination
	printf(" done!");
	printf("\n\nTo recover the original file, simply concatenate the file slices sequentially\nin binary mode.\n");

	return 0;
}
